function defines(...)
	local str = ''
	for i,v in ipairs{...} do
		str = str .. '#define '..tostring(v)..'\n'
	end
	return str
end

defaultVertexCode = [[
void main() {
	//preserve all the stuff our render pathway could send our way
	//this should be the same for just about all shaders
	gl_FrontColor = gl_Color;				//primary color of the sprite
	gl_TexCoord[0] = gl_MultiTexCoord0;		//texunit 0 is associated with the spritetex & bumptex, in world coords
	gl_TexCoord[1] = gl_MultiTexCoord1;		//texunit 1 is associated with the lighttex & screentex, in screen coords
	gl_Position = ftransform();
}
]]
fragmentShaderUniforms = [[

//the sprite base
uniform sampler2D spritetex;

//the lightmap accumed from a fbo just before the frame began
uniform sampler2D lighttex;

//the sprite bumpmap, in tangent-space coordinates.
//pushed back to 3 so our non-shader pathway can just use units 0 and 1
uniform sampler2D bumptex;

//the last screen tex texture accumed into a fbo
//useful for feedback effects (image flow, etc)
//and more efficient than using the current fbo
//(which would require a texture upload ... which requires no change in texture bindings actually 
uniform sampler2D lastscreentex;

//the current screen texture
//if you use this then the render pathway notices
//and from that it takes care to upload the current framebuffer to a separate texture
//and pass it in by here
//so using this texture inevitably will cause an extra texcopy (in fbo so it wont go too slow)
//
// AVOID USING THIS UNIT IF AT ALL POSSIBLE!
// it is bound to unit4 for a reason - so if there are any old 4-unit cards they will miss out on it
//	do NOT use it for world graphics - the number of uploads kills performance
// it is ONLY good for a single sprite that takes in (refraction?) the contents behind it but not in front of it
//
// A CHEAP ALTERNATIVE to the same effect is to make your sprite postFBO and have it use lastscreentex
//	postFBO sprites are rendered after the fbo accumulation, and therefore dont get the recurrence that usually occurs
//	 this will render your shader on top of all else, but it will have an accurate display (of the frame before) of what is below it
// (good for explosions and the likes)
//
uniform sampler2D currentscreentex;

uniform float time;
]];
defaultFragmentCode = fragmentShaderUniforms .. [[
void main() {
	vec4 spriteColor = texture2D(spritetex, gl_TexCoord[0].st);
	vec4 lightColor = texture2D(lighttex, gl_TexCoord[1].st);
	gl_FragColor = spriteColor * lightColor;
}
]];
--[[
awesome
but now i have a new issue (that the bug was hiding)
we draw our invis guy with distorted background
but then he gets fed into the FBO
and the next frame his invis isn't distorting the background, but instead his last disorted image
thus image flow - there's our watercolor effect - but how do we keep the original pred invis?

well, we'd have to have a separate copy of the FBO without the sprite
that'd mean marking certain sprites as "draw me after the FBO is done"
and that'd mean using depth testing again (not a problem ... i think ... well, except the whole blending thing)

i did use a separate fbo to fix the lighting/blend sort issue
can i do that here as well?

i'm going to throw in a "draw last" flag
for both the HUD (to avoid it getting mixed into the FBO)
and for the FBO effects (to keep them from being recursive - unless so desired)

this one relies on the FBOTEX #define.  no defaults, so if you don't use it right then it'll complain
--]]
distortFragmentCode = fragmentShaderUniforms .. [[
void main() {
	vec4 spriteColor = texture2D(spritetex, gl_TexCoord[0].st);
	vec4 lightColor = texture2D(lighttex, gl_TexCoord[1].st);
	vec4 bumpColor = texture2D(bumptex, gl_TexCoord[0].st);
	vec4 ofsScreenColor = texture2D(FBOTEX, gl_TexCoord[1].st + (bumpColor.st * 2. - 1.) * -.02);
	gl_FragColor.rgb = ofsScreenColor.rgb;
	gl_FragColor.a = spriteColor.a;
}
]];
-- invis means read from screentex (slowdown) while advect (image flow) doesn't
-- there is a cheap hack though for fast invis (so long as it is rendered on top of everything)
-- you just use the advection texture with the postFBO flag set
invisibilityFragmentCode = defines('FBOTEX currentscreentex') .. distortFragmentCode
advectionFragmentCode = defines('FBOTEX lastscreentex') .. distortFragmentCode

bumpMapFragmentCode = fragmentShaderUniforms .. [[
void main() {
	vec4 spriteColor = texture2D(spritetex, gl_TexCoord[0].st);
	vec4 lightColor = texture2D(lighttex, gl_TexCoord[1].st);
	vec4 bumpColor = texture2D(bumptex, gl_TexCoord[0].st);
	//true bump-mapping requires we accum lighting scaled by the dot3 of the surface with the light direction vector
	//that means either mod'ing the light accum pathway in the renderer (more correct)
	//or passing some hack vars as lightpos's to sum across (faster? less changes)
	gl_FragColor.rgb = spriteColor.rgb * lightColor.rgb * (bumpColor.r * 2. - 1.);
	gl_FragColor.a = spriteColor.a;
}
]];

function renderShader(_args)
	local args = union({}, _args)	-- copy the args
	if not args.vertexCode and not args.vertexFile then args.vertexCode = defaultVertexCode end
	-- a fragment has to at least be provided
	if not args.uniforms then args.uniforms = {} end	-- so we can mod them
	union(args.uniforms, {
		spritetex = 0;
		lighttex = 1;
		bumptex = 2;
		lastscreentex = 3;
		currentscreentex = 4;
	})
	return createShader(args)
end

-- should I pre-create a few of these?  think anyone who is hoarding handles will mind?
-- meh, they can always not include this file ... or can they?
-- maybe i should make a separate 'shader presets' file
-- ... once we find a shader hoard at least =)
